<?php

namespace Drupal\entity_extender\Plugin\EntityTypeExtender;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\RevisionableInterface;
use Drupal\entity_extender\Plugin\EntityExtender\EntityLifecycleInterface;

/**
 * The extended storage handler for entities.
 */
trait EntityHandlerTraitStorage {

  use EntityHandlerTrait;

  /**
   * {@inheritdoc}
   */
  public function create(array $values = []) {
    $entity = parent::create($values);
    $this->lifecycle($entity, static function (EntityLifecycleInterface $extender): void { $extender->onCreate(); });

    return $entity;
  }

  /**
   * {@inheritdoc}
   */
  public function createRevision(RevisionableInterface $entity, $default = TRUE, $keep_untranslatable_fields = NULL) {
    $revision = parent::createRevision($entity, $default, $keep_untranslatable_fields);
    $this->lifecycle($entity, static function (EntityLifecycleInterface $extender): void { $extender->onCreateRevision(); });

    return $revision;
  }

  /**
   * {@inheritdoc}
   */
  protected function invokeFieldMethod($method, ContentEntityInterface $entity, ...$args): array {
    switch ($method) {
      /* @see \Drupal\Core\Entity\ContentEntityStorageBase::deleteRevision() */
      case 'deleteRevision':
        $this->lifecycle($entity, static function (EntityLifecycleInterface $extender): void { $extender->onPreDeleteRevision(); });
        break;
    }

    return parent::invokeFieldMethod($method, $entity, ...$args);
  }

  /**
   * {@inheritdoc}
   */
  protected function invokeHook($hook, EntityInterface $entity): void {
    parent::invokeHook($hook, $entity);

    switch ($hook) {
      /* @see \Drupal\Core\Entity\EntityStorageBase::doPreSave() */
      case 'presave':
        $this->lifecycle($entity, static function (EntityLifecycleInterface $extender): void { $extender->onPreSave(); });
        break;

      /* @see \Drupal\Core\Entity\EntityStorageBase::doPostSave() */
      case 'insert':
        $this->lifecycle($entity, static function (EntityLifecycleInterface $extender): void { $extender->onInsert(); });
        break;

      /* @see \Drupal\Core\Entity\EntityStorageBase::doPostSave() */
      case 'update':
        $this->lifecycle($entity, static function (EntityLifecycleInterface $extender): void { $extender->onUpdate(); });
        break;

      /* @see \Drupal\Core\Entity\EntityStorageBase::delete() */
      case 'predelete':
        $this->lifecycle($entity, static function (EntityLifecycleInterface $extender): void { $extender->onPreDelete(); });
        break;

      /* @see \Drupal\Core\Entity\EntityStorageBase::delete() */
      case 'delete':
        $this->lifecycle($entity, static function (EntityLifecycleInterface $extender): void { $extender->onPostDelete(); });
        break;

      /* @see \Drupal\Core\Entity\ContentEntityStorageBase::deleteRevision() */
      case 'revision_delete':
        $this->lifecycle($entity, static function (EntityLifecycleInterface $extender): void { $extender->onPostDeleteRevision(); });
        break;
    }
  }

  /**
   * Runs the lifecycle handler.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   * @param \Closure $handler
   *   The callback to run the lifecycle.
   */
  private function lifecycle(EntityInterface $entity, \Closure $handler): void {
    // We must ensure the extender is created from scratch during each
    // lifecycle event. This is especially crucial during entity creation/edit
    // from UI, so the `presave`/`update`/`insert` has updated values.
    static::undoMemo($entity);

    if ($extender = $this->entityExtenderManager->getExtenderIfMemberOf(EntityLifecycleInterface::class, $entity)) {
      \assert($extender instanceof EntityLifecycleInterface);
      $handler($extender);
    }
  }

}
